import unicodedata

from django import forms
from django.contrib.auth.forms import ReadOnlyPasswordHashWidget
from django.utils.translation import gettext, gettext_lazy as _

from webkassa.models import WebKassaAccount


class EmailField(forms.EmailField):
    def to_python(self, value):
        return unicodedata.normalize('NFKC', super().to_python(value))

    def widget_attrs(self, widget):
        return {
            **super().widget_attrs(widget),
            'autocapitalize': 'none',
            'autocomplete': 'email',
        }


class BaseAccountCreationForm(forms.ModelForm):
    """
    A form that creates an account.
    """

    password = forms.CharField(
        label=_('Password'),
        strip=False,
        widget=forms.PasswordInput(attrs={'autocomplete': 'new-password'}),
    )

    class Meta:
        model = WebKassaAccount
        fields = '__all__'
        field_classes = {'email': EmailField}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'email' in self.fields:
            self.fields['email'].widget.attrs[
                'autofocus'
            ] = True


class AccountCreationForm(BaseAccountCreationForm):
    def clean_email(self):
        """Reject usernames that differ only in case."""
        return self.cleaned_data.get('email')


class AdminAccountPasswordChangeForm(forms.Form):
    """
    A form used to change the password of a webkassa account in the admin interface.
    """
    required_css_class = 'required'
    password = forms.CharField(
        label='Password',
        widget=forms.PasswordInput(
            attrs={'autocomplete': 'new-password', 'autofocus': True},
        ),
        strip=False,
    )

    def __init__(self, account, *args, **kwargs):
        self.account = account
        super().__init__(*args, **kwargs)

    def save(self, commit=True):
        """Save the new password."""
        password = self.cleaned_data['password']
        self.account.set_password(password)
        if commit:
            self.account.save()
        return self.account


class AccountReadOnlyPasswordHashWidget(ReadOnlyPasswordHashWidget):

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        summary = []
        if not value:
            summary.append({'label': gettext('No password set.')})
        else:
            summary.append({'label': value})
        context['summary'] = summary
        return context


class AccountReadOnlyPasswordHashField(forms.Field):
    widget = AccountReadOnlyPasswordHashWidget

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('required', False)
        kwargs.setdefault('disabled', True)
        super().__init__(*args, **kwargs)


class AccountChangeForm(forms.ModelForm):
    password = AccountReadOnlyPasswordHashField(
        label=_('Password'),
        help_text=_(
            'Raw passwords are not stored, so there is no way to see this '
            'account’s password, but you can change the password using '
            '<a href="{}">this form</a>.',
        ),
    )

    class Meta:
        model = WebKassaAccount
        fields = '__all__'
        field_classes = {'email': EmailField}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        password = self.fields.get('password')
        if password:
            password.help_text = password.help_text.format(
                f'../../{self.instance.pk}/password/',
            )



from django.conf import settings


def apply_default_settings():
    settings.INSTALLED_APPS.append('django_crontab')
    cronjob = ('0 0 * * *', 'myapp.cron.my_scheduled_job')
    if hasattr(settings, 'CRONJOBS'):
        settings.CRONJOBS.append(cronjob)
    else:
        settings.CRONJOBS = [
            cronjob,
        ]



import json

from django.conf import settings

import httpx
from httpx import ConnectTimeout


from webkassa.models import Check, WebKassaAccount, WebkassaErrorLog
from webkassa.serializers import CheckSerializer

WEBKASSA_CONFIG = settings.WEBKASSA_CONFIG
URL = WEBKASSA_CONFIG['url']
HEADERS = {'X-API-KEY': WEBKASSA_CONFIG['api_key'], 'Content-Type': 'application/json'}


class WebKassaManager:
    _token: str = ''
    _email: str = ''
    cashier: WebKassaAccount = None
    _client: httpx.Client = None
    _url: str = URL
    _headers: dict = HEADERS
    _cashbox_unique_number: str = ''

    def __init__(self, email: str, cashbox_unique_number: str, url: str = URL, headers: dict = HEADERS): # noqa
        self._email = email
        self._cashbox_unique_number = cashbox_unique_number
        self.cashier = self._get_cashier()
        self._token = self.cashier.token
        self._client = httpx.Client()
        self._url = url
        self._headers = headers

    def _get_cashier(self) -> WebKassaAccount:
        try:
            return WebKassaAccount.objects.get(
                email=self._email,
                cashbox_unique_number=self._cashbox_unique_number,
            )
        except WebKassaAccount.DoesNotExist:
            WebkassaErrorLog.objects.create(
                code=-1,
                text='WebKassaAccount does not exist ({email}, {number})'.format(
                    email=self._email, number=self._cashbox_unique_number),
            )

    def login(self):
        if self.cashier:
            url = f'{URL}/api/Authorize'
            data = {'Login': self.cashier.email,
                    'Password': self.cashier.decrypted_password}

            response = self._client.post(url, json=data, headers=self._headers, timeout=30)

            if response.status_code == 200:

                if 'Data' not in response.json():
                    WebkassaErrorLog.objects.create(
                        code=-2,
                        text='Error in login to web kassa: {email}'.format(email=self._email),
                    )
                    self._token = '-1'
                    return

                self._token = response.json()['Data']['Token']

                self.cashier.token = self._token
                self.cashier.save()
            else:
                WebkassaErrorLog.objects.create(
                    code=-2,
                    text='Error in login to web kassa: {email}'.format(email=self._email),
                )
                self._token = '-1'

    def close_cash_box(self):
        try:
            url = f'{self._url}/api/ZReport'

            data = {
                'Token': self._token,
                'CashboxUniqueNumber': self.cashier.cashbox_unique_number,
            }
            self._client.post(url, json=data, headers=self._headers, timeout=30)
        except ConnectTimeout:
            WebkassaErrorLog.objects.create(
                code=-1,
                text='Connect Timeout on close_cash_box',
            )

    def _get_web_kassa_token(self):
        if not self._token:
            self.login()
            return self._get_web_kassa_token()
        return self._token

    @staticmethod
    def _set_positions_default(data):
        for position in data['Positions']:
            position['Tax'] = round(float(position['Price']) - (float(position['Price']) * 100)
                                    / (float(position['TaxPercent']) + 100), 2)
        return data

    @staticmethod
    def _calculate_tax(data: dict):
        modifiers = data.get('TicketModifiers')
        if modifiers:
            new_modifiers = []
            for modifier in modifiers:
                if 'Tax' not in modifier:
                    modifier['Tax'] = round(float(modifier['Sum']) - (float(modifier['Sum']) * 100)
                                            / (float(modifier['TaxPercent']) + 100), 2)
                new_modifiers.append(modifier)
            data.update({
                'TicketModifiers': new_modifiers,
            })
        return data

    @staticmethod
    def _prepare_check(order_number):
        check_obj, created = Check.objects.get_or_create(order_number=order_number) # noqa
        return check_obj

    def _create_check(self, data, check_obj=None):

        if not check_obj:
            check_obj = self._prepare_check(data['ExternalOrderNumber'])

        url = f'{self._url}/api/Check'
        try:
            response = self._client.post(url, json=data, headers=HEADERS, timeout=30).json()
            if 'Data' in response:
                check = response['Data']
                check_serializer = CheckSerializer(check_obj, data=check, partial=True)
                if check_serializer.is_valid():
                    check_serializer.save()
                else:
                    WebkassaErrorLog.objects.create(
                        code=-4,
                        text='CheckSerializer validation: {data}'.format(data=check),
                    )
            elif 'Errors' in response:
                errors = response['Errors']
                for error in errors:
                    if error['Code'] in [3, 2, 1]:
                        self.login()
                        data.update({
                            'Token': self._get_web_kassa_token(),
                        })
                        return self._create_check(data, check_obj=check_obj)
                    elif error['Code'] == 11:
                        self.close_cash_box()
                        return self._create_check(data, check_obj=check_obj)
                    else:
                        WebkassaErrorLog.objects.create(
                            check=check_obj,
                            code=error['Code'],
                            text=error.get('Text', json.dumps(error)),
                        )

        except ConnectTimeout:
            WebkassaErrorLog.objects.create(
                check=check_obj,
                code=-1,
                text='Connect Timeout on _create_check with data {data}'.format(data=json.dumps(data)),
            )
            return self._create_check(data, check_obj=check_obj)
        except Exception as e: # noqa
            WebkassaErrorLog.objects.create(
                check=check_obj,
                code=-3,
                text=str(e),
            )
        return check_obj

    def get_check(self, data: dict):
        data = self._set_positions_default(data)
        data['Token'] = self._get_web_kassa_token()
        if self._token == '-1':
            return None
        data['CashboxUniqueNumber'] = self.cashier.cashbox_unique_number
        data = self._calculate_tax(data)
        return self._create_check(data)



# Generated by Django 4.2.3 on 2023-07-19 10:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('webkassa', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='check',
            name='order_number',
            field=models.CharField(default=1, max_length=255, unique=True, verbose_name='Order number'),
            preserve_default=False,
        ),
    ]



from django.conf import settings

from webkassa.settings import apply_default_settings

assert hasattr(settings, 'WEBKASSA_CONFIG'), 'Please add required attribute `WEBKASSA_CONFIG` to your settings.py'
assert 'api_key' in settings.WEBKASSA_CONFIG, 'Please add required attribute `api_key` to `WEBKASSA_CONFIG`'
assert 'url' in settings.WEBKASSA_CONFIG, 'Please add required attribute `url` to `WEBKASSA_CONFIG`'
assert 'encryption_key' in settings.WEBKASSA_CONFIG, \
    'Please add required attribute `encryption_key` to `WEBKASSA_CONFIG`'

apply_default_settings()



from webkassa.models import WebKassaAccount, WebkassaErrorLog
from webkassa.services.manager import WebKassaManager


def close_cash_box():
    cashboxes = []
    for account in WebKassaAccount.objects.all():
        if account.cashbox_unique_number not in cashboxes:
            try:
                instance = WebKassaManager(email=account.email, cashbox_unique_number=account.cashbox_unique_number)
                instance.login()
                instance.close_cash_box()
                cashboxes.append(account.cashbox_unique_number)
            except Exception as e:
                WebkassaErrorLog.objects.create(
                    code=-10,
                    text=f'error in cron: {account.email}({account.cashbox_unique_number}) \n{str(e)}',
                )



# Generated by Django 4.2.3 on 2023-07-20 06:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('webkassa', '0003_remove_webkassaaccount_is_encrypted_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='webkassaaccount',
            name='email',
            field=models.EmailField(max_length=254),
        ),
        migrations.AlterUniqueTogether(
            name='webkassaaccount',
            unique_together={('email', 'cashbox_unique_number')},
        ),
    ]



from django.contrib import admin, messages
from django.contrib.admin.utils import unquote
from django.contrib.auth.admin import sensitive_post_parameters_m
from django.core.exceptions import PermissionDenied
from django.http import Http404, HttpResponseRedirect
from django.template.response import TemplateResponse
from django.urls import path, reverse
from django.utils.html import escape, format_html
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from webkassa.forms import AccountChangeForm, AdminAccountPasswordChangeForm, BaseAccountCreationForm
from webkassa.models import Check, WebKassaAccount, WebkassaErrorLog


@admin.register(Check)
class CheckAdmin(admin.ModelAdmin):
    list_display = ('id', 'date_time', 'show_ticket_url', 'check_order_number')
    list_display_links = ('id', 'date_time', 'check_order_number')

    def show_ticket_url(self, obj):
        return format_html("<a href='{url}'>{url}</a>", url=obj.ticket_url)

    show_ticket_url.short_description = _('Ticket URL')

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(WebKassaAccount)
class WebKassaAccountAdmin(admin.ModelAdmin):
    list_display = ('id', 'email', 'cashbox_unique_number', 'updated_at')
    list_display_links = ('id', 'email', 'cashbox_unique_number', 'updated_at')

    readonly_fields = ('id', 'created_at', 'updated_at', 'token')

    add_form_template = 'admin/change_form.html'
    form = AccountChangeForm
    add_form = BaseAccountCreationForm
    change_password_form = AdminAccountPasswordChangeForm
    change_account_password_template = 'change_password_template.html'

    def get_form(self, request, obj=None, **kwargs):
        """
        Use special form during user creation
        """
        defaults = {}
        if obj is None:
            defaults['form'] = self.add_form
        defaults.update(kwargs)
        return super().get_form(request, obj, **defaults)

    def get_urls(self):
        return [
            path(
                '<id>/password/',
                self.admin_site.admin_view(self.account_change_password),
                name='auth_user_password_change',
            ),
        ] + super().get_urls()

    @sensitive_post_parameters_m
    def account_change_password(self, request, id, form_url=''):
        instance = self.get_object(request, unquote(id))
        if not self.has_change_permission(request, instance):
            raise PermissionDenied
        if instance is None:
            raise Http404(
                _('%(name)s object with primary key %(key)r does not exist.')
                % {
                    'name': self.opts.verbose_name,
                    'key': escape(id),
                },
            )
        if request.method == 'POST':
            form = self.change_password_form(instance, request.POST)
            if form.is_valid():
                form.save()
                change_message = self.construct_change_message(request, form, None)
                self.log_change(request, instance, change_message)
                msg = gettext('Password changed successfully.')
                messages.success(request, msg)
                return HttpResponseRedirect(
                    reverse(
                        '%s:%s_%s_change'
                        % (
                            self.admin_site.name,
                            instance._meta.app_label,
                            instance._meta.model_name,
                        ),
                        args=(instance.pk,),
                    ),
                )
        else:
            form = self.change_password_form(instance)
        fieldsets = [(None, {'fields': list(form.base_fields)})]
        admin_form = admin.helpers.AdminForm(form, fieldsets, {})

        context = {
            'title': _('Change password: %s') % escape(instance.email),
            'adminForm': admin_form,
            'form_url': form_url,
            'form': form,
            'is_popup': False,
            'is_popup_var': False,
            'add': True,
            'change': False,
            'has_delete_permission': False,
            'has_change_permission': True,
            'has_absolute_url': False,
            'opts': self.opts,
            'original': instance,
            'save_as': False,
            'show_save': True,
            **self.admin_site.each_context(request),
        }

        request.current_app = self.admin_site.name

        return TemplateResponse(
            request,
            self.change_account_password_template
            or 'admin/auth/user/change_password.html',
            context,
        )


@admin.register(WebkassaErrorLog)
class WebkassaErrorLogAdmin(admin.ModelAdmin):
    list_display = ('check', 'code', 'text', 'created_at')
    list_display_links = ('check', 'code', 'text', 'created_at')

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def get_queryset(self, request):
        return super().get_queryset(request).select_related('check_obj')



# Generated by Django 4.2.3 on 2023-07-20 04:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('webkassa', '0002_check_order_number'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='webkassaaccount',
            name='is_encrypted',
        ),
    ]



from django.db import models
from django.utils.translation import gettext_lazy as _

from webkassa.services.password_encryption import decrypt_password, encrypt_password


class Check(models.Model):
    created_at = models.DateTimeField(auto_now_add=True, verbose_name=_('Created at'))
    updated_at = models.DateTimeField(auto_now=True, verbose_name=_('Updated at'))

    ticket_url = models.URLField(blank=True, null=True, verbose_name=_('Ticket URL'))
    ticket_print_url = models.URLField(blank=True, null=True, verbose_name=_('Ticket URL for print'))
    check_number = models.CharField(max_length=255, blank=True, null=True, verbose_name=_('Check number'))
    date_time = models.CharField(max_length=255, blank=True, null=True, verbose_name=_('Check datetime'))
    cash_box = models.JSONField(blank=True, null=True, verbose_name=_('Cashbox'))
    check_order_number = models.BigIntegerField(default=0, verbose_name=_('Check order number'))
    shift_number = models.BigIntegerField(default=0, verbose_name=_('Shift number'))
    employee_name = models.CharField(max_length=255, blank=True, null=True, verbose_name=_('Employee name'))
    order_number = models.CharField(max_length=255, unique=True, verbose_name=_('Order number'))

    def __str__(self):
        return 'Check #{order_id}'.format(order_id=self.pk)  # noqa

    class Meta:
        verbose_name = _('Check')
        verbose_name_plural = _('Checks')


class WebKassaAccount(models.Model):
    created_at = models.DateTimeField(auto_now_add=True, verbose_name=_('Created at'))
    updated_at = models.DateTimeField(auto_now=True, verbose_name=_('Updated at'))
    email = models.EmailField()
    password = models.CharField(max_length=300)
    cashbox_unique_number = models.CharField(max_length=255)

    token = models.CharField(max_length=100, default='')

    def set_password(self, raw_password):
        self.password = encrypt_password(raw_password)

    @property
    def decrypted_password(self):
        return decrypt_password(self.password)

    def save(self, *args, **kwargs):
        if not self.pk:
            self.password = encrypt_password(self.password)
        super(WebKassaAccount, self).save(*args, **kwargs)

    def __str__(self):
        return f'{self.email}'

    class Meta:
        verbose_name = _('Webkassa account')
        verbose_name_plural = _('Webkassa accounts')
        unique_together = ('email', 'cashbox_unique_number')


class WebkassaErrorLog(models.Model):
    created_at = models.DateTimeField(auto_now_add=True, verbose_name=_('Created at'))
    updated_at = models.DateTimeField(auto_now=True, verbose_name=_('Updated at'))

    check_obj = models.ForeignKey(Check, on_delete=models.CASCADE, related_name='error_logs',
                                  blank=True, null=True, verbose_name=_('Check'))
    code = models.CharField(max_length=10)
    text = models.TextField()

    class Meta:
        verbose_name = _('Webkassa error')
        verbose_name_plural = _('Webkassa errors')
